import unittest


class TruthTableTestCase(unittest.TestCase):
    """
    Tests:
        - create output by_enumerating_output_sequence
    """

    @unittest.SkipTest
    def test_first(self):
        self.fail()
